using System;
using iNTERFACEWARE.Chameleon;
using System.Reflection;
using System.Windows.Forms;  // we use the single threaded windows timer.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
// NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// iNTERFACEWARE BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CON-
// NECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

namespace Hl7Service
{
    class ClientModule
    {
        private bool runningAsService;
        private LlpClient client;
        private Timer tickTimer;
        private ClientState state;
        private System.IO.StreamWriter logStream;

        private static string APP_LOG_FILE = "hl7log.txt";
        private static ushort APP_REMOTE_PORT = 5123;
        private static string APP_REMOTE_HOST = "localhost";
        private static int APP_MAX_MESSAGE_LOG_LENGTH = 2048;  // maximum size of message to display
        private static int APP_TICK_INTERVAL = 1000; // interval of our tick timer.
        private static int APP_ACK_TIMEOUT = 120; // seconds

        public ClientModule(bool isService)
        {
            runningAsService = isService;
            if (runningAsService)
            {
                logStream = System.IO.File.AppendText(ExecutableDirectory() + APP_LOG_FILE);
            }
        }

        private string ExecutableDirectory()
        {
            return System.IO.Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + "\\";
        }

        public void StartService()
        {
            try
            {
                // We need to instantiate the LLP Client here since it needs to be in the thread
                // that runs the message loop.
                client = new LlpClient();

                client.MessageReceived += new LlpClient.MessageReceivedEvent(Client_MessageReceived);
                client.Closed += new LlpClient.ClosedEvent(Client_Closed);
                client.ErrorReceived += new LlpClient.ErrorReceivedEvent(Client_ErrorReceived);
                client.DataIgnored += new LlpClient.DataIgnoredEvent(Client_DataIgnored);
                client.Connected += new LlpClient.ConnectedEvent(Client_Connected);

                tickTimer = new Timer();
                tickTimer.Interval = APP_TICK_INTERVAL;
                tickTimer.Tick += new EventHandler(TickTimer_Tick);
                tickTimer.Start();

                state = new ClientState();
                
                // Run the message loop.  This is blocking.  If you wish to add functionality
                // to the start up of the service, be sure to add it *before* this call.
                iNTERFACEWARE.Chameleon.LlpClient.RunMessageLoop();
                // Cleaning up after the message loop has stopped.
                client.Disconnect();
            }
            catch (Exception ex)
            {
                Log("Error starting service: " + ex.Message.ToString());
            }
        }

        public void StopService()
        {
            try
            {
                // Stop the messsage loop.  Because this is on another thread we cannot
                // do anything else.
                client.StopMessageLoop();
            }
            catch (Exception ex)
            {
                Log(ex.Message.ToString());
            }
        }      

        private void ConnectToRemoteHost()
        {
            Log("Connecting to port " + APP_REMOTE_PORT + " on " + APP_REMOTE_HOST);
            client.Connect(APP_REMOTE_HOST, APP_REMOTE_PORT);
        }

        private void PollAndSendMessage()
        {
            Log("Polling for message to send...");

            if (state.FetchMessage())
            {
                Log("found id=" + state.MessageId);
                // We have a message to send
                state.SetAckTimeOut(APP_ACK_TIMEOUT);
                // Sending the message over the network can always result in an exception since the remote host
                // can disconnect at any time.  The application behavior is safe since it should simply time out
                // and then resend the message.
                TransmitMessageWithLog();
            }
            else
            {
                Log("No message to send.");
            }
        }

        private void TransmitMessageWithLog()
        {
            Log("Sending:");
            Log(PrettyPrintMessage(state.Message));
            client.SendMessage(state.Message);
            Log("Sent!");
        }

        private void DoConnectedAction()
        {
            switch (state.CurrentState)
            {
            case Hl7State.NEED_TO_POLL_APP:
                PollAndSendMessage();
                return;
            case Hl7State.WAITING_FOR_ACK:
                if (state.IsTimedOut())
                {
                    Log("Timed out waiting for ACK for " + state.MessageId);
                    // We need to resend the message - we reset the ACK timeout
                    // TODO - we're just resending the message here - other possible implementations would be to disconnect
                    // before trying to resend.

                    // TODO - if we get many resends in a row it would probably make sense to keep a count and then
                    // send an email or pager alert.
                    state.SetAckTimeOut(APP_ACK_TIMEOUT);
                    Log("Resending...");
                    // and resend the message
                    TransmitMessageWithLog();
                }
                return;
           };
        }

        private void TickTimer_Tick(object sender, EventArgs eArgs)
        {
            if (!client.IsConnected)
            {
                ConnectToRemoteHost();
            }
            else
            {
                DoConnectedAction();
            }
        }

        private void Client_MessageReceived(object Sender, System.EventArgs e, string message)
        {
            Log("Received an ack");
            Log(PrettyPrintMessage(message));
 
            switch (state.CurrentState)
            {
                case Hl7State.WAITING_FOR_ACK:
                    // TODO - we'll assume that the ACK is well formed and the message control ID matches the one we were expecting
                    // We could parse that out with Chameleon if we wanted to be very paranoid.  In a synchronous exchange it should
                    // be okay...
                    // The following command is *VERY IMPORTANT* it marks the message as being sent - the ClientState code should
                    // be written in a manner to never consider a message sent until this application gives confirmation.
                    Log(state.MessageId + " marked as sent.");
                    state.ConfirmMessageSend();
                    return;
                case Hl7State.NEED_TO_POLL_APP:
                    // TODO page alert
                    Log("WARNING!  We were not expecting an ACK.");
                    return;
            }             
        }

        private void Client_Closed(object sender, System.EventArgs e)
        {
            Log("Connection Closed");
        }

        private void Client_ErrorReceived(object Sender, System.EventArgs e, string description, int code)
        {
            Log("Error Received : " + PrettyPrintMessage(description));
        }
   	
        private void Client_DataIgnored(object sender, System.EventArgs e, string data)	
        {		
            Log("Ignored Data received: ");
            Log(PrettyPrintMessage(data));
        }

        private void Client_Connected(object sender, System.EventArgs e)
        {
            Log("Connection established with host");
        }

        // This routine converts \r --> \r\n and truncates the message if necessary so that
        // HL7 messages can be seen in the log
        private string PrettyPrintMessage(string message)
        {
            return client.MakeMessagePrintable(message, "\r\n", APP_MAX_MESSAGE_LOG_LENGTH);
        }

        private void Log(string output)
        {
            // TODO some type of Log Purging will be required in a real application.
            string logLine = System.DateTime.Now.ToString() + " " + output;
                
            if (runningAsService)
            {
                logStream.WriteLine(logLine);
                logStream.Flush();
            }
            else
            {
                System.Console.WriteLine(logLine);
            }
        }      
    }
}
