using System;
using System.Threading;
using System.Runtime.InteropServices;
using System.Reflection;

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
// NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// iNTERFACEWARE BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CON-
// NECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

namespace ConsoleUtil
{
	/// <summary>
	/// Simple utility class to catch console control events (ie CTRL-C) in C#. 
	/// Calls SetConsoleCtrlHandler() in Win32 API.  This is used as part of this example
    /// to provide a means of stopping the service when it is running in command line mode.
	/// </summary>
	public class ConsoleCtrl : IDisposable
	{
		/// <summary>
		/// The event that occurred. 
		/// </summary>
		public enum ConsoleEvent
		{
			CTRL_C = 0,		// From wincom.h
			CTRL_BREAK = 1,
			CTRL_CLOSE = 2,
			CTRL_LOGOFF = 5,
			CTRL_SHUTDOWN = 6
		}

		/// <summary>
		/// Handler to be called when a console event occurs.
		/// </summary>
		public delegate void ControlEventHandler(ConsoleEvent consoleEvent);

		/// <summary>
		/// Event fired when a console event occurs
		/// </summary>
		public event ControlEventHandler ControlEvent;

		ControlEventHandler eventHandler;

		/// <summary>
		/// Create a new instance.
		/// </summary>
		public ConsoleCtrl()
		{
			// save this to a private var so the GC doesn't collect it...
			eventHandler = new ControlEventHandler(Handler);
			SetConsoleCtrlHandler(eventHandler, true);
		}

		~ConsoleCtrl()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
		}

		private void Dispose(bool disposing)
		{
			if (disposing)
			{
				GC.SuppressFinalize(this);
			}
			if (eventHandler != null)
			{
				SetConsoleCtrlHandler(eventHandler, false);
				eventHandler = null;
			}
		}

		private void Handler(ConsoleEvent consoleEvent)
		{
            if (ControlEvent != null)
            {
                ControlEvent(consoleEvent);
            }
		}

		[DllImport("kernel32.dll")]
		static extern bool SetConsoleCtrlHandler(ControlEventHandler e, bool add);
	}
}